//Copyright (C) cocosimulator.org 2014
//generated by AmsterCHEM's COM CAPE-OPEN Wizard

#pragma once
#include "CapeOpenBase.h"
#include "Water.h"
#include "COException.h"

//WaterPackage.h - definitions of CAPE-OPEN WaterPackage class

class WaterPackage :
	public ICapeThermoCompounds,
	public ICapeThermoPhases,
	public ICapeThermoMaterialContext,
	public ICapeThermoPropertyRoutine,
	public ICapeThermoUniversalConstant,
	public ICapeThermoEquilibriumRoutine,
	public ICapeUtilities,
	/*BASE_WaterPackage_INSERT_LOCATION*/
	public CAPEOPENBASE {

	CBStrConst temperature,pressure,fraction,phaseFraction,mole,mass,liquid,vapor,enthalpy,entropy;	
	COMSmartPtr<ICapeThermoMaterial> contextMat;
	CVariant compIDs,compFormulae,compBoilTemps,compMolWts,compCasnos;
	CVariant phaseLabels,phaseAggState,phaseKeyComps;
	vector<wstring> phases;
	bool compoundsChecked; //a typical property package would store compound indices instead of ok. In this case, the list of compounds must be {water}; any subset is invalid (empty)
	//cached vectors, defined as members for efficiency (to avoid heap allocations)
	vector<CBStr> names;
	vector<BSTR> bases;
	vector<double> values;
	CVariant scalar;

public:

	//constructor
	WaterPackage();
	//destructor
	~WaterPackage() {
		//call terminate, in case it has not yet been called by the COSE
		Terminate();
	}

	IMPLEMENT_IDISPATCH

	//overrides:
	virtual void *InternalQueryInterface(REFIID riid) {
		//interfaces implemented by this class:
		if (riid==IID_ICapeThermoCompounds) return (ICapeThermoCompounds*)this;
		if (riid==IID_ICapeThermoPhases) return (ICapeThermoPhases*)this;
		if (riid==IID_ICapeThermoMaterialContext) return (ICapeThermoMaterialContext*)this;
		if (riid==IID_ICapeThermoPropertyRoutine) return (ICapeThermoPropertyRoutine*)this;
		if (riid==IID_ICapeThermoUniversalConstant) return (ICapeThermoUniversalConstant*)this;
		if (riid==IID_ICapeThermoEquilibriumRoutine) return (ICapeThermoEquilibriumRoutine*)this;
		if (riid==IID_ICapeUtilities) return (ICapeUtilities*)this;
		/*INTERFACE_WaterPackage_INSERT_LOCATION*/
		return CAPEOPENBASE::InternalQueryInterface(riid); //return base class interfaces
	}

	//utility functions

	double GetSinglePhaseScalarProp(BSTR prop,BSTR phase,BSTR basis) {
		//check before hand that we have a context material
		ASSERT(contextMat!=NULL,"no context material");
		CVariant propVal;
		HRESULT hr;
		wstring err;
		hr=contextMat->GetSinglePhaseProp(prop,phase,basis,propVal.OutputArgument());
		if (FAILED(hr)) {
			err=L"Failed to get ";
			err+=prop;
			err+=L" for phase ";
			err+=phase;
			err+=L": ";
			err+=CO_Error(contextMat,hr);
			throw COException(err);
		}
		if (!propVal.CheckArray(VT_R8,err)) {
			wstring error;
			error=L"Invalid value for ";
			error+=prop;
			error+=L" for phase ";
			error+=phase;
			error+=L": ";
			error+=err;
			throw COException(error);
		}
		if (propVal.GetCount()!=1) {
			err=L"Invalid value for ";
			err+=prop;
			err+=L" for phase ";
			err+=phase;
			err+=L": scalar expected";
			throw COException(err);
		}
		return propVal.GetDoubleAt(0);
	}

	double GetOverallScalarProp(BSTR prop,BSTR basis) {
		//check before hand that we have a context material
		ASSERT(contextMat!=NULL,"no context material");
		wstring err;
		CVariant propVal;
		HRESULT hr;
		hr=contextMat->GetOverallProp(prop,basis,propVal.OutputArgument());
		if (FAILED(hr)) {
			err=L"Failed to get ";
			err+=prop;
			err+=L" for overall phase: ";
			err+=CO_Error(contextMat,hr);
			throw COException(err);
		}
		if (!propVal.CheckArray(VT_R8,err)) {
			wstring error;
			error=L"Invalid value for ";
			error+=prop;
			error+=L" for overall phase: ";
			error+=err;
			throw COException(error);
		}
		if (propVal.GetCount()!=1) {
			err=L"Invalid value for ";
			err+=prop;
			err+=L" for overall phase: ";
			throw COException(err);
		}
		return propVal.GetDoubleAt(0);
	}

	void SetSinglePhaseScalarProp(double val,BSTR prop,BSTR phase,BSTR basis) {
		//check before hand that we have a context material
		ASSERT(contextMat!=NULL,"no context material");
		HRESULT hr;
		CVariant value;
		value.MakeArray(1,VT_R8);
		value.SetDoubleAt(0,val);
		hr=contextMat->SetSinglePhaseProp(prop,phase,basis,value.Value());
		if (FAILED(hr)) {
			wstring err;
			err=L"Failed to set ";
			err+=prop;
			err+=L" for phase ";
			err+=phase;
			err+=L": ";
			err+=CO_Error(contextMat,hr);
			throw COException(err);
		}
	}

	void SetOverallScalarProp(double val,BSTR prop,BSTR basis) {
		//check before hand that we have a context material
		ASSERT(contextMat!=NULL,"no context material");
		wstring ws;
		HRESULT hr;
		CVariant value;
		value.MakeArray(1,VT_R8);
		value.SetDoubleAt(0,val);
		hr=contextMat->SetOverallProp(prop,basis,value.Value());
		if (FAILED(hr)) {
			wstring err;
			err=L"Failed to set ";
			err+=prop;
			err+=L" for overall phase: ";
			err+=CO_Error(contextMat,hr);
			throw COException(err);
		}
	}

	void CheckCompounds() {
		if (!compoundsChecked) {
			wstring error;
			//check the compounds on the context MO
			// a typical property package should analyze the compound list on the MO and store for example a list of indices
			// as a subset of compounds may be present, or compounds may be in the different order. As we expose only 
			// "water" the only valid compound list is {"water"}
			ASSERT(contextMat!=NULL,"no context material");
			COMSmartPtr<ICapeThermoCompounds> comp;
			HRESULT hr=contextMat->QueryInterface(IID_ICapeThermoCompounds,(LPVOID*)&comp);
			if (FAILED(hr)) {
				throw COException(L"Context material does not implement ICapeThermoCompounds");
			}
			CVariant compIds,formulae,names,boilTemps,molwts,casnos;
			hr=comp->GetCompoundList(compIds.OutputArgument(),formulae.OutputArgument(),names.OutputArgument(),boilTemps.OutputArgument(),molwts.OutputArgument(),casnos.OutputArgument());
			if (FAILED(hr)) {
				throw COException(L"Failed to obtain list of compounds from context material: "+CO_Error(comp,hr));
			}
			if (!compIds.CheckArray(VT_BSTR,error)) {
				throw COException(L"Invalid list of compounds from context material: "+error);
			}
			if (compIds.GetCount()!=1) {
				throw COException(L"Invalid list of compounds from context material: expected {\"water\"}");
			}
			if (compIds.GetStringAt(0)!=L"Water") {
				throw COException(L"Invalid list of compounds from context material: expected {\"water\"}");
			}
			compoundsChecked=true; //no need to check again until SetMaterial is called
		}
	}

	void GetTPX(BSTR phaseLabel,double &T,double &P) {
		//in our case, X must always be {1}
		wstring error;
		ASSERT(phaseLabel,"check valid phase label before calling this function");
		CheckCompounds();
		CVariant composition;
		HRESULT hr=contextMat->GetTPFraction(phaseLabel,&T,&P,composition.OutputArgument());
		if (FAILED(hr)) {
			error=L"Failed to obtain T,P,X for ";
			error+=phaseLabel;
			error+=L" from context material: ";
			error+=CO_Error(contextMat,hr);
			throw COException(error);
		}
		if (isMissing(T)) {
			throw COException(L"Missing value for temperature");
		}
		if (isMissing(P)) {
			throw COException(L"Missing value for pressure");
		}
		if (!composition.CheckArray(VT_R8,error)) {
			throw COException(L"Invalid value for composition: "+error);
		}
		if (composition.GetCount()!=1) {
			throw COException(L"Invalid value for composition: invalid number of values");
		}
		if (fabs(composition.GetDoubleAt(0)-1.0)>1e-3) {
			throw COException(L"Invalid value for composition: not normalized");
		}
	}

	void GetOverallTPX(double &T,double &P) {
		//in our case, X must always be {1}
		wstring error;
		CheckCompounds();
		CVariant composition;
		HRESULT hr=contextMat->GetOverallTPFraction(&T,&P,composition.OutputArgument());
		if (FAILED(hr)) {
			throw COException(L"Failed to obtain overall T,P,X from context material: "+CO_Error(contextMat,hr));
		}
		if (isMissing(T)) {
			throw COException(L"Missing value for temperature");
		}
		if (isMissing(P)) {
			throw COException(L"Missing value for pressure");
		}
		if (!composition.CheckArray(VT_R8,error)) {
			throw COException(L"Invalid value for composition: "+error);
		}
		if (composition.GetCount()!=1) {
			throw COException(L"Invalid value for composition: invalid number of values");
		}
		if (fabs(composition.GetDoubleAt(0)-1.0)>1e-3) {
			throw COException(L"Invalid value for composition: not normalized");
		}
	}

	void CheckOverallX() {
		//in our case, X must always be {1}
		wstring error;
		CheckCompounds();
		CVariant composition;
		HRESULT hr=contextMat->GetOverallProp(fraction,mole,composition.OutputArgument());
		if (FAILED(hr)) {
			throw COException(L"Failed to obtain overall composition from context material: "+CO_Error(contextMat,hr));
		}
		if (!composition.CheckArray(VT_R8,error)) {
			throw COException(L"Invalid value for composition: "+error);
		}
		if (composition.GetCount()!=1) {
			throw COException(L"Invalid value for composition: invalid number of values");
		}
		if (fabs(composition.GetDoubleAt(0)-1.0)>1e-3) {
			throw COException(L"Invalid value for composition: not normalized");
		}
	}

	//phase lookups

	int GetPhaseIndex(const wchar_t *phaseLabel) {
		if (!phaseLabel) throw COException(L"Empty phase label",ECapeInvalidArgumentHR);
		for (size_t i=0;i<phases.size();i++) {
			if (SameString(phaseLabel,phases[i].c_str())) {
				return (int)i;
			}
		}
		wstring error;
		error=L"Invalid/unsupported phase \"";
		error+=phaseLabel;
		error+=L'"';
		throw COException(error);		
	}

	void GetPhasePair(VARIANT phasePair,int &index1,int &index2) {
		CVariant phases(phasePair,FALSE);
		wstring error;
		if (!phases.CheckArray(VT_BSTR,error)) {
			throw COException(L"Invalid phase pair: "+error,ECapeInvalidArgumentHR);
		}
		if (phases.GetCount()!=2) {
			throw COException(L"Invalid phase pair: expected two phases");
		}
		index1=GetPhaseIndex(phases.GetStringAt(0));
		index2=GetPhaseIndex(phases.GetStringAt(1));
		if (index1==index2) {
			throw COException(L"Invalid phase pair: phases are the same");
		}
	}

	//Phase equilibrium type

	enum EqType {
	    //most commonly used flashes, and the ones supported by this object
	    TP,	//pressure, temperature
	    TVF,	//temperature, vapor fraction
	    PVF,	//pressure, vapor fraction
	    PH,	//pressure, enthalpy
	    TH,	//temperature, enthalpy
	    PS,	//pressure, entropy
	    TS, 	//temperature, entropy
	    HVF,  //enthalpy, vapor fraction
	    SVF   //entropy, vapor fraction
	};

	void GetEquilibriumType(VARIANT specification1, VARIANT specification2, BSTR solutionType, EqType &type, BSTR &VFBasis,BOOL &vfVapor) {
		//this routine does most of the work of CheckEquilibriumSpec and part of the work of CalcEquilibrium
		int i;
		HRESULT hr;
		BOOL haveVap,haveLiq;
		wstring err;
		BOOL haveT,haveP,haveVF,haveH,haveS;
		VFBasis=NULL; //fill in if vapor fraction is part of flash type
		vfVapor=TRUE; //set to false for liquid fraction spec
		haveT=haveP=haveVF=haveH=haveS=FALSE;
		//check if we have a context material
		if (contextMat==NULL) {
			throw COException(L"Context material not set");
		}
		//get list of present phases (we only support flashes that can result vapor and liquid)
		CVariant phaseList,phaseStatus;
		hr=contextMat->GetPresentPhases(phaseList.OutputArgument(),phaseStatus.OutputArgument());
		if (FAILED(hr)) {
			throw COException(L"Failed to get list of present phases from context material: "+CO_Error(contextMat,hr));
		}
		if (!phaseList.CheckArray(VT_BSTR,err)) {
			throw COException(L"Invalid list of present phases from context material: "+err);
		}
		if (phaseList.GetCount()!=2) {
			throw COException(L"Only flashes that can result vapor and liquid are supported");
		}
		haveVap=haveLiq=FALSE;
		for (i=0; i<2; i++) {
			BSTR label=phaseList.GetStringAt(i);
			if (SameString(label,vapor)) {
				haveVap=TRUE;
			} else if (SameString(label,liquid)) {
				haveLiq=TRUE;
			}
		}
		if ((!haveVap)||(!haveLiq)) {
			throw COException(L"Only flashes that can result vapor and liquid are supported");
		}
		//check both specifications, to see what they are
		for (i=0; i<2; i++) {
			CVariant spec((i)?specification2:specification1,FALSE);
			try {
				//check spec format
				if (!spec.CheckArray(VT_BSTR,err)) {
					throw COException(err);
				}
				if ((spec.GetCount()<3)||(spec.GetCount()>4)) {
					throw COException(L"expected 3 or 4 strings");
				}
				//get property, basis and phase
				CBStr prop=spec.GetStringAt(0);
				CBStr basis=spec.GetStringAt(1);
				CBStr phase=spec.GetStringAt(2);
				//which property?
				if (SameString(prop,temperature)) {
					//T spec, only for overall
					if (!SameString(phase,L"Overall")) {
						throw COException(L"temperature specification only supported for overall phase");
					}
					if (!SameString(basis,NULL)) {
						throw COException(L"unsupported basis for temperature");
					}
					if (haveT) {
						throw COException(L"Multiple temperature flash conditions");
					}
					haveT=TRUE;
				} else if (SameString(prop,pressure)) {
					//P spec, only for overall
					if (!SameString(phase,L"Overall")) {
						throw COException(L"pressure specification only supported for overall phase");
					}
					if (!SameString(basis,NULL)) {
						throw COException(L"unsupported basis for pressure");
					}
					if (haveP) {
						throw COException(L"Multiple pressure flash conditions");
					}
					haveP=TRUE;
				} else if (SameString(prop,phaseFraction)) {
					//phase fraction spec
					int index=GetPhaseIndex(phase);
					vfVapor=(index==0);
					if (SameString(basis,mass)) {
						VFBasis=mass;
					} else if (SameString(basis,mole)) {
						VFBasis=mole;
					} else {
						throw COException(L"unsupported basis for phase fraction specification");
					}
					if (haveVF) {
						throw COException(L"Multiple phase fraction conditions");
					}
					haveVF=TRUE;
				} else if (SameString(prop,L"enthalpy")) {
					//H spec, only for overall
					if (!SameString(phase,L"Overall")) {
						throw COException(L"enthalpy specification only supported for overall phase");
					}
					if (haveH) {
						throw COException(L"Multiple enthalpy conditions");
					}
					haveH=TRUE;
				} else if (SameString(prop,L"entropy")) {
					//S spec, only for overall
					if (!SameString(phase,L"Overall")) {
						throw COException(L"entropy specification only supported for overall phase");
					}
					if (haveS) {
						throw COException(L"Multiple entropy conditions");
					}
					haveS=TRUE;
				} else {
					err=L"Unsupported property ";
					err+=(prop)?prop:L"NULL";
					throw COException(err);
				}
				//we do not support a flash spec for a certain compound, just for the whole mixture of only water
				if (spec.GetCount()==4) {
					if (!SameString(spec.GetStringAt(3),NULL)) {
						throw COException(L"compound specification not supported");
					}
				}
			} catch (COException &ex) {
				wstringenh error;
				error.printf(L"Invalid flash specification %d: ",i+1);
				ex.Prepend(error);
				throw ex;
			}
		}
		//check flash type
		if (haveT) {
			if (haveP) {
				type=TP;
			} else if (haveVF) {
				type=TVF;
			} else if (haveH) {
				type=TH;
			} else if (haveS) {
				type=TS;
			} else {
				throw COException(L"Unsupported combination of flash specification variables");
			}
		} else if (haveP) {
			if (haveVF) {
				type=PVF;
			} else if (haveH) {
				type=PH;
			} else if (haveS) {
				type=PS;
			} else {
				throw COException(L"Unsupported combination of flash specification variables");
			}
		} else if (haveVF) {
			if (haveH) {
				type=HVF;
			} else if (haveS) {
				type=SVF;
			} else {
				throw COException(L"Unsupported combination of flash specification variables");
			}
		} else {
			throw COException(L"Unsupported combination of flash specification variables");
		}
		//check solution type
		if ((!SameString(solutionType,NULL))&&(!SameString(solutionType,L"Unspecified"))) {
			//a solution type is specified
			if (SameString(solutionType,L"Normal")) {
				//normal, only ok if VF is specified
				if (!haveVF) {
					throw COException(L"Solution type \"normal\" is only supported for phase fraction flashes");
				}
			} else if (SameString(solutionType,L"Retrograde")) {
				throw COException(L"Retrograde flashes are not supported");
			} else {
				err=L"Invalid or unsupported solution type: ";
				err+=solutionType;
				throw COException(err);
			}
		}
	}


	//ICapeThermoCompounds Interface

	STDMETHOD(GetCompoundConstant)(/*[in]*/ VARIANT props,/*[in]*/ VARIANT compIds,/*[out,retval] */ VARIANT *compoundConstant);
	STDMETHOD(GetCompoundList)(/*[in,out]*/ VARIANT* compIds,/*[in,out]*/ VARIANT* formulae,/*[in,out]*/ VARIANT* names,/*[in,out]*/ VARIANT* boilTemps,/*[in,out]*/ VARIANT* molwts,/*[in,out]*/ VARIANT* casnos);
	STDMETHOD(GetConstPropList)(/*[out,retval] */ VARIANT *constPropList);
	STDMETHOD(GetNumCompounds)(/*[out,retval] */ long int *numCompounds);
	STDMETHOD(GetPDependentProperty)(/*[in]*/ VARIANT props,/*[in]*/ double pressure,/*[in]*/ VARIANT compIds,/*[in,out]*/ VARIANT* propVals);
	STDMETHOD(GetPDependentPropList)(/*[out,retval] */ VARIANT *pDependentPropList);
	STDMETHOD(GetTDependentProperty)(/*[in]*/ VARIANT props,/*[in]*/ double temperature,/*[in]*/ VARIANT compIds,/*[in,out]*/ VARIANT* propVals);
	STDMETHOD(GetTDependentPropList)(/*[out,retval] */ VARIANT *tDependentPropList);

	//ICapeThermoPhases Interface

	STDMETHOD(GetNumPhases)(/*[out,retval] */ long int *numPhases);
	STDMETHOD(GetPhaseInfo)(/*[in]*/ BSTR phaseLabel,/*[in]*/ BSTR phaseAttribute,/*[out,retval] */ VARIANT *phaseInfo);
	STDMETHOD(GetPhaseList)(/*[in,out]*/ VARIANT* phaseLabels,/*[in,out]*/ VARIANT* stateOfAggregation,/*[in,out]*/ VARIANT* keyCompoundId);

	//ICapeThermoMaterialContext Interface

	STDMETHOD(SetMaterial)(/*[in]*/ LPDISPATCH material);
	STDMETHOD(UnsetMaterial)();

	//ICapeThermoPropertyRoutine Interface

	STDMETHOD(CalcAndGetLnPhi)(/*[in]*/ BSTR phaseLabel,/*[in]*/ double temperature,/*[in]*/ double pressure,/*[in]*/ VARIANT moleNumbers,/*[in]*/ int fFlags,/*[in,out]*/ VARIANT* lnPhi,/*[in,out]*/ VARIANT* lnPhiDT,/*[in,out]*/ VARIANT* lnPhiDP,/*[in,out]*/ VARIANT* lnPhiDn);
	STDMETHOD(CalcSinglePhaseProp)(/*[in]*/ VARIANT props,/*[in]*/ BSTR phaseLabel);
	STDMETHOD(CalcTwoPhaseProp)(/*[in]*/ VARIANT props,/*[in]*/ VARIANT phaseLabels);
	STDMETHOD(CheckSinglePhasePropSpec)(/*[in]*/ BSTR property,/*[in]*/ BSTR phaseLabel,/*[out,retval] */ VARIANT_BOOL *checkSinglePhasePropSpecResult);
	STDMETHOD(CheckTwoPhasePropSpec)(/*[in]*/ BSTR property,/*[in]*/ VARIANT phaseLabels,/*[out,retval] */ VARIANT_BOOL *checkTwoPhasePropSpecResult);
	STDMETHOD(GetSinglePhasePropList)(/*[out,retval] */ VARIANT *singlePhasePropList);
	STDMETHOD(GetTwoPhasePropList)(/*[out,retval] */ VARIANT *twoPhasePropList);

	//ICapeThermoUniversalConstant Interface

	STDMETHOD(GetUniversalConstant)(/*[in]*/ BSTR constantId,/*[out,retval] */ VARIANT *universalConstant);
	STDMETHOD(GetUniversalConstantList)(/*[out,retval] */ VARIANT *universalConstantList);

	//ICapeThermoEquilibriumRoutine Interface

	STDMETHOD(CalcEquilibrium)(/*[in]*/ VARIANT specification1,/*[in]*/ VARIANT specification2,/*[in]*/ BSTR solutionType);
	STDMETHOD(CheckEquilibriumSpec)(/*[in]*/ VARIANT specification1,/*[in]*/ VARIANT specification2,/*[in]*/ BSTR solutionType,/*[out,retval] */ VARIANT_BOOL *checkEquilibriumSpecResult);

	//ICapeUtilities Interface

	STDMETHOD(get_parameters)(/*[out,retval] */ LPDISPATCH *parameters);
	STDMETHOD(put_simulationContext)(/*[in]*/ LPDISPATCH simulationContext);
	STDMETHOD(Initialize)();
	STDMETHOD(Terminate)();
	STDMETHOD(Edit)();

	/*CODE_WaterPackage_INSERT_LOCATION*/

};

